import os
import shutil
import unittest

from unittest.mock import patch

from secator.config import CONFIG
from secator.output_types import Vulnerability
from secator.utils_test import FIXTURES_DIR, clear_modules


class TestTemplate(unittest.TestCase):

	def setUp(self):
		self.template_dir = CONFIG.dirs.templates
		self.custom_task_path = self.template_dir / 'ls.py'
		self.writeable_file = self.template_dir / 'test.txt'
		self.custom_workflow_path = self.template_dir / 'ls.yml'
		shutil.copy(f'{FIXTURES_DIR}/ls.py', self.custom_task_path)
		shutil.copy(f'{FIXTURES_DIR}/ls.yml', self.custom_workflow_path)
		self.writeable_file.touch()
		os.chmod(self.writeable_file, 0o007)
		self.expected_vuln = Vulnerability(
			name='World-writeable path',
			severity='high',
			confidence='high',
			provider='ls',
			matched_at=f'{str(self.writeable_file)}',
			_source='ls',
		)
		clear_modules()
		self.maxDiff = None

	def tearDown(self):
		self.custom_task_path.unlink()
		self.custom_workflow_path.unlink()
		self.writeable_file.unlink()

	def test_external_task(self):
		from secator.tasks import ls
		task = ls(str(self.template_dir))
		task.run()
		findings = task.findings
		self.assertEqual(len(findings), 1)
		self.assertTrue(self.expected_vuln == Vulnerability.load(findings[0].toDict()))

	def test_external_workflow(self):
		from secator.cli import ALL_WORKFLOWS
		from secator.runners import Workflow
		ls_workflow = [w for w in ALL_WORKFLOWS if w.name == 'ls'][0]
		self.assertIsNotNone(ls_workflow)
		workflow = Workflow(ls_workflow, inputs=[str(self.template_dir)])
		workflow.run()
		findings = workflow.findings
		self.assertEqual(len(findings), 1)
		vuln = [r for r in findings if r._type == 'vulnerability'][0]
		self.assertTrue(self.expected_vuln == Vulnerability.load(vuln.toDict()))


class TestTemplateLoader(unittest.TestCase):

	def setUp(self):
		self.task_config = {
			'type': 'task',
			'name': 'nuclei'
		}
		self.workflow_config = {
			'type': 'workflow',
			'name': 'test',
			'tasks': {
				'nuclei': {'opt_1': 'test1'},
				'nmap': {'opt_2': 'test2'},
				'_group': {
					'httpx': {'opt_3': 'test3'},
					'nuclei/host': {'opt_4': 'test4'}
				},
				'nuclei/network': {'opt_4': 'test4_new', 'opt_5': 'test5'}
			}
		}
		self.scan_config = {
			'type': 'scan',
			'name': 'test',
			'workflows': {
				'test/1': {},
				'test/2': {}
			}
		}

	def test_extract_tasks_and_supported_opts_task(self):
		# Test different configurations
		from secator.tasks import nuclei
		from secator.template import TemplateLoader
		loader = TemplateLoader(input=self.task_config)
		tasks = loader._extract_tasks()
		self.assertEqual(list(tasks.keys()), ['nuclei'])
		self.assertEqual(tasks['nuclei']['class'], nuclei)

	def test_extract_tasks_and_supported_opts_workflow(self):
		# Test different configurations
		from secator.runners import Task
		from secator.template import TemplateLoader
		loader = TemplateLoader(input=self.workflow_config)
		tasks = loader._extract_tasks()
		self.assertEqual(list(tasks.keys()), ['nuclei', 'nmap', 'httpx', 'nuclei/host', 'nuclei/network'])
		for task, task_config in tasks.items():
			self.assertEqual(task.split('/')[0], task_config['name'])
			self.assertEqual(task_config['class'], Task.get_task_class(task_config['name']))
		self.assertEqual(tasks['nuclei']['opts'], {'opt_1': 'test1'})
		self.assertEqual(tasks['nmap']['opts'], {'opt_2': 'test2'})
		self.assertEqual(tasks['httpx']['opts'], {'opt_3': 'test3'})
		self.assertEqual(tasks['nuclei/host']['opts'], {'opt_4': 'test4'})
		self.assertEqual(tasks['nuclei/network']['opts'], {'opt_4': 'test4_new', 'opt_5': 'test5'})

	@patch('secator.template.TemplateLoader._load_from_name')
	def test_extract_tasks_and_supported_opts_scan(self, mock_load_from_name):
		from secator.template import TemplateLoader
		mock_load_from_name.return_value = self.workflow_config
		loader_wf = TemplateLoader(input=self.workflow_config)
		wf_tasks = loader_wf._extract_tasks()
		loader = TemplateLoader(input=self.scan_config)
		tasks = loader._extract_tasks()
		expected_tasks = {}
		for workflow in ['test/1', 'test/2']:
			for task, task_conf in wf_tasks.items():
				expected_tasks[f'{workflow}/{task}'] = task_conf
		self.assertEqual(tasks, expected_tasks)



from datetime import date, datetime
import json
from pathlib import PosixPath
from secator.output_types import OUTPUT_TYPES


class DataclassEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'toDict'):
            return obj.toDict()
        elif isinstance(obj, PosixPath):
            return str(obj)
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        else:
            return json.JSONEncoder.default(self, obj)


def get_output_cls(type):
    try:
        return [cls for cls in OUTPUT_TYPES if cls.get_name() == type][0]
    except IndexError:
        return None


def dataclass_decoder(obj):
    if '_type' in obj:
        output_cls = get_output_cls(obj['_type'])
        if output_cls:
            return output_cls.load(obj)
    return obj


def dumps_dataclass(obj, indent=None):
    return json.dumps(obj, cls=DataclassEncoder, indent=indent)


def loads_dataclass(obj):
    return json.loads(obj, object_hook=dataclass_decoder)



import os
from pathlib import Path
from subprocess import call, DEVNULL
from typing import Dict, List
from typing_extensions import Annotated, Self

import requests
import yaml
from dotenv import find_dotenv, load_dotenv
from dotmap import DotMap
from pydantic import AfterValidator, BaseModel, model_validator, ValidationError

from secator.rich import console, console_stdout

load_dotenv(find_dotenv(usecwd=True), override=False)

Directory = Annotated[Path, AfterValidator(lambda v: v.expanduser())]
StrExpandHome = Annotated[str, AfterValidator(lambda v: v.replace('~', str(Path.home())))]

ROOT_FOLDER = Path(__file__).parent.parent
LIB_FOLDER = ROOT_FOLDER / 'secator'
CONFIGS_FOLDER = LIB_FOLDER / 'configs'
DATA_FOLDER = os.environ.get('SECATOR_DIRS_DATA') or str(Path.home() / '.secator')


class StrictModel(BaseModel, extra='forbid'):
	pass


class Directories(StrictModel):
	bin: Directory = Path.home() / '.local' / 'bin'
	share: Directory = Path.home() / '.local' / 'share'
	data: Directory = Path(DATA_FOLDER)
	templates: Directory = ''
	reports: Directory = ''
	wordlists: Directory = ''
	cves: Directory = ''
	payloads: Directory = ''
	performance: Directory = ''
	revshells: Directory = ''
	celery: Directory = ''
	celery_data: Directory = ''
	celery_results: Directory = ''

	@model_validator(mode='after')
	def set_default_folders(self) -> Self:
		"""Set folders to be relative to the data folders if they are unspecified in config."""
		for folder in ['templates', 'reports', 'wordlists', 'cves', 'payloads', 'performance', 'revshells', 'celery', 'celery_data', 'celery_results']:  # noqa: E501
			rel_target = '/'.join(folder.split('_'))
			val = getattr(self, folder) or self.data / rel_target
			setattr(self, folder, val)
		return self


class Debug(StrictModel):
	level: int = 0
	component: str = ''


class Celery(StrictModel):
	broker_url: str = 'filesystem://'
	broker_pool_limit: int = 10
	broker_connection_timeout: float = 4.0
	broker_visibility_timeout: int = 3600
	broker_transport_options: str = ""
	override_default_logging: bool = True
	result_backend: StrExpandHome = ''
	result_backend_transport_options: str = ""
	result_expires: int = 86400  # 1 day
	task_acks_late: bool = False
	task_send_sent_event: bool = False
	task_reject_on_worker_lost: bool = False
	worker_max_tasks_per_child: int = 20
	worker_prefetch_multiplier: int = 1
	worker_send_task_events: bool = False
	worker_kill_after_task: bool = False
	worker_kill_after_idle_seconds: int = -1


class Cli(StrictModel):
	github_token: str = os.environ.get('GITHUB_TOKEN', '')
	record: bool = False
	stdin_timeout: int = 1000


class Runners(StrictModel):
	input_chunk_size: int = 100
	progress_update_frequency: int = 20
	stat_update_frequency: int = 20
	backend_update_frequency: int = 5
	poll_frequency: int = 5
	skip_cve_search: bool = False
	skip_exploit_search: bool = False
	skip_cve_low_confidence: bool = False
	remove_duplicates: bool = False


class Security(StrictModel):
	allow_local_file_access: bool = True
	auto_install_commands: bool = True
	force_source_install: bool = False


class HTTP(StrictModel):
	socks5_proxy: str = 'socks5://127.0.0.1:9050'
	http_proxy: str = 'https://127.0.0.1:9080'
	store_responses: bool = False
	response_max_size_bytes: int = 100000  # 100MB
	proxychains_command: str = 'proxychains'
	freeproxy_timeout: int = 1


class Tasks(StrictModel):
	exporters: List[str] = ['json', 'csv', 'txt']


class Workflows(StrictModel):
	exporters: List[str] = ['json', 'csv', 'txt']


class Scans(StrictModel):
	exporters: List[str] = ['json', 'csv', 'txt']


class Payloads(StrictModel):
	templates: Dict[str, str] = {
		'lse': 'https://github.com/diego-treitos/linux-smart-enumeration/releases/latest/download/lse.sh',
		'linpeas': 'https://github.com/carlospolop/PEASS-ng/releases/latest/download/linpeas.sh',
		'sudo_killer': 'https://github.com/TH3xACE/SUDO_KILLER/archive/refs/heads/V3.zip'
	}


class Wordlists(StrictModel):
	defaults: Dict[str, str] = {'http': 'bo0m_fuzz', 'dns': 'combined_subdomains'}
	templates: Dict[str, str] = {
		'bo0m_fuzz': 'https://raw.githubusercontent.com/Bo0oM/fuzz.txt/master/fuzz.txt',
		'combined_subdomains': 'https://raw.githubusercontent.com/danielmiessler/SecLists/master/Discovery/DNS/combined_subdomains.txt',  # noqa: E501
		'directory_list_small': 'https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Discovery/Web-Content/directory-list-2.3-small.txt',  # noqa: E501
	}
	lists: Dict[str, List[str]] = {}


class GoogleDriveAddon(StrictModel):
	enabled: bool = False
	drive_parent_folder_id: str = ''
	credentials_path: str = ''


class GoogleCloudStorageAddon(StrictModel):
	enabled: bool = False
	bucket_name: str = ''
	credentials_path: str = ''


class WorkerAddon(StrictModel):
	enabled: bool = False


class MongodbAddon(StrictModel):
	enabled: bool = False
	url: str = 'mongodb://localhost'
	update_frequency: int = 60
	max_pool_size: int = 10
	server_selection_timeout_ms: int = 5000


class Addons(StrictModel):
	gdrive: GoogleDriveAddon = GoogleDriveAddon()
	gcs: GoogleCloudStorageAddon = GoogleCloudStorageAddon()
	worker: WorkerAddon = WorkerAddon()
	mongodb: MongodbAddon = MongodbAddon()


class SecatorConfig(StrictModel):
	dirs: Directories = Directories()
	debug: Debug = Debug()
	celery: Celery = Celery()
	cli: Cli = Cli()
	runners: Runners = Runners()
	http: HTTP = HTTP()
	tasks: Tasks = Tasks()
	workflows: Workflows = Workflows()
	scans: Scans = Scans()
	payloads: Payloads = Payloads()
	wordlists: Wordlists = Wordlists()
	addons: Addons = Addons()
	security: Security = Security()
	offline_mode: bool = False


class Config(DotMap):
	"""Config class.

	Examples:
	>>> config = Config.parse()  						   # get default config.
	>>> config = Config.parse({'dirs': {'data': '/tmp/'})  # get custom config (from dict).
	>>> config = Config.parse(path='/path/to/config.yml')  # get custom config (from YAML file).
	>>> config.print() 									   # print config without defaults.
	>>> config.print(partial=False)  					   # print full config.
	>>> config.set('addons.gdrive.enabled', False)         # set value in config.
	>>> config.save()									   # save config back to disk.
	"""

	_error = False

	def get(self, key=None, print=True):
		"""Retrieve a value from the configuration using a dotted path.

		Args:
			key (str | None): Dotted key path.
			print (bool): Print the resulting value.

		Returns:
			Any: value at key.
		"""
		value = self
		if key:
			for part in key.split('.'):
				value = value[part]
		if value is None:
			console.print(f'[bold red]Key {key} does not exist.[/]')
			return None
		if print:
			if key:
				yaml_str = Config.dump(DotMap({key: value}), partial=False)
			else:
				yaml_str = Config.dump(self, partial=False)
			Config.print_yaml(yaml_str)
		return value

	def set(self, key, value, set_partial=True):
		"""Set a value in the configuration using a dotted path.

		Args:
			key (str | None): Dotted key path.
			value (Any): Value.
			set_partial (bool): Set in partial config.
		"""
		# Get existing value
		existing_value = self.get(key, print=False)

		# Convert dotted key path to the corresponding uppercase key used in _keymap
		map_key = key.upper().replace('.', '_')

		# Check if map key exists
		if map_key not in self._keymap:
			console.print(f'[bold red]Key "{key}" not found in config keymap[/].')
			return

		# Traverse to the second last key to handle the setting correctly
		target = self
		partial = self._partial
		for part in self._keymap[map_key][:-1]:
			target = target[part]
			partial = partial[part]

		# Set the value on the final part of the path
		final_key = self._keymap[map_key][-1]

		# Try to convert value to expected type
		try:
			if isinstance(existing_value, list):
				if isinstance(value, str):
					if value.startswith('[') and value.endswith(']'):
						value = value[1:-1]
					if ',' in value:
						value = [c.strip() for c in value.split(',')]
					elif value:
						value = [value]
					else:
						value = []
			elif isinstance(existing_value, dict):
				if isinstance(value, str):
					if value.startswith('{') and value.endswith('}'):
						import json
						value = json.loads(value)
			elif isinstance(existing_value, bool):
				if isinstance(value, str):
					value = value.lower() in ("true", "1", "t")
				elif isinstance(value, (int, float)):
					value = True if value == 1 else False
			elif isinstance(existing_value, int):
				value = int(value)
			elif isinstance(existing_value, float):
				value = float(value)
			elif isinstance(existing_value, Path):
				value = Path(value)
		except ValueError:
			# from secator.utils import debug
			# debug(f'Could not cast value {value} to expected type {type(existing_value).__name__}: {str(e)}', sub='config')
			pass
		finally:
			target[final_key] = value
			if set_partial:
				partial[final_key] = value

	def save(self, target_path: Path = None, partial=True):
		"""Save config as YAML on disk.

		Args:
			target_path (Path | None): If passed, saves the config to this path.
			partial (bool): Save partial config.
		"""
		if not target_path:
			if not self._path:
				return
			target_path = self._path
		with target_path.open('w') as f:
			f.write(Config.dump(self, partial=partial))
		self._path = target_path

	def print(self, partial=True):
		"""Print config.

		Args:
			partial (bool): Print partial config only.
		"""
		yaml_str = self.dump(self, partial=partial)
		yaml_str = f'# {self._path}\n\n{yaml_str}' if self._path and partial else yaml_str
		Config.print_yaml(yaml_str)

	@staticmethod
	def parse(data: dict = {}, path: Path = None, print_errors: bool = True):
		"""Parse config.

		Args:
			data (dict): Config data.
			path (Path | None): Path to YAML config.
			print_errors (bool): Print validation errors to console.

		Returns:
			Config: instance of Config object.
			None: if the config was not loaded properly or there are validation errors.
		"""
		# Load YAML file
		if path:
			data = Config.read_yaml(path)

		# Load data
		config = Config.load(SecatorConfig, data, print_errors=print_errors)
		valid = config is not None
		if not valid:
			return None

		# Set extras
		config.set_extras(data, path)

		# Override config values with environment variables
		config.apply_env_overrides(print_errors=print_errors)

		# Validate config
		config.validate(print_errors=print_errors)

		return config

	def validate(self, print_errors=True):
		"""Validate config."""
		return Config.load(
			SecatorConfig,
			data=self._partial.toDict(),
			print_errors=print_errors)

	def set_extras(self, original_data, original_path):
		"""Set extra useful values in config.

		Args:
			original_data (data): Original dict data.
			original_path (pathlib.Path): Original YAML path.
			valid (bool): Boolean indicating if config is valid or not.
		"""
		self._path = original_path
		self._partial = Config(original_data)
		self._keymap = Config.build_key_map(self)

		# HACK: set default result_backend if unset
		if not self.celery.result_backend:
			self.celery.result_backend = f'file://{self.dirs.celery_results}'

	@staticmethod
	def load(schema, data: dict = {}, print_errors=True):
		"""Validate a config using Pydantic.

		Args:
			schema (pydantic.Schema): Pydantic schema.
			data (dict): Input data.
			print_errors (bool): Print validation errors.

		Returns:
			Config|None: instance of Config object or None if invalid.
		"""
		try:
			return Config(schema(**data).model_dump())
		except ValidationError as e:
			if print_errors:
				error_str = str(e).replace('\n', '\n  ')
				console.print(f'[bold red]:x: {error_str}')
			return None

	@staticmethod
	def read_yaml(yaml_path):
		"""Read YAML from path.

		Args:
			yaml_path (Path): path to yaml config.

		Returns:
			dict: Loaded data.
		"""
		if not yaml_path.exists():
			console.print(f'[bold red]Config not found: {yaml_path}.[/]')
			return {}
		try:
			with yaml_path.open('r') as f:
				data = yaml.load(f.read(), Loader=yaml.Loader)
				return data or {}
		except yaml.YAMLError as e:
			console.print(f'[bold red]:x: Error loading {yaml_path} {str(e)}')
			return {}

	@staticmethod
	def print_yaml(string):
		"""Print YAML string using rich.

		Args:
			string (str): YAML string.
		"""
		from rich.syntax import Syntax
		data = Syntax(string, 'yaml', theme='ansi-dark', padding=0, background_color='default')
		console_stdout.print(data)

	@staticmethod
	def dump(config, partial=True):
		"""Safe dump config as yaml:
		- `Path`, `PosixPath` and `WindowsPath` objects are translated to strings.
		- Home directory in paths is replaced with the tilde '~'.

		Returns:
			str: YAML dump.
		"""
		import yaml
		from pathlib import Path, PosixPath, WindowsPath

		# Get home dir
		home = str(Path.home())

		# Custom dumper to add line breaks between items and a path representer to translate paths to strings
		class LineBreakDumper(yaml.SafeDumper):
			def write_line_break(self, data=None):
				super().write_line_break(data)
				if len(self.indents) == 1:
					super().write_line_break()

		def posix_path_representer(dumper, data):
			path = str(data)
			if path.startswith(home):
				path = path.replace(home, '~')
			return dumper.represent_scalar('tag:yaml.org,2002:str', path)

		LineBreakDumper.add_representer(str, posix_path_representer)
		LineBreakDumper.add_representer(Path, posix_path_representer)
		LineBreakDumper.add_representer(PosixPath, posix_path_representer)
		LineBreakDumper.add_representer(WindowsPath, posix_path_representer)

		# Get data dict
		data = config.toDict()

		# HACK: Replace home dir in result_backend
		if isinstance(config, Config):
			data['celery']['result_backend'] = data['celery']['result_backend'].replace(home, '~')
			del data['_path']
			if partial:
				data = data['_partial']
			else:
				del data['_partial']

		data = {k: v for k, v in data.items() if not k.startswith('_')}
		return yaml.dump(data, Dumper=LineBreakDumper, sort_keys=False)

	@staticmethod
	def build_key_map(config, base_path=[]):
		key_map = {}
		for key, value in config.items():
			if key.startswith('_'):  # ignore
				continue
			current_path = base_path + [key]
			if isinstance(value, dict):
				key_map.update(Config.build_key_map(value, current_path))
			else:
				key_map['_'.join(current_path).upper()] = current_path
		return key_map

	def apply_env_overrides(self, print_errors=True):
		"""Override config values from environment variables."""
		prefix = "SECATOR_"
		for var in os.environ:
			if var.startswith(prefix):
				key = var[len(prefix):]  # remove prefix
				if key in self._keymap:
					path = '.'.join(k.lower() for k in self._keymap[key])
					value = os.environ[var]
					self.set(path, value, set_partial=False)
					if not self.validate(print_errors=False) and print_errors:
						console.print(f'[bold red]{var} (override failed)[/]')
				# elif print_errors:
				# 	console.print(f'[bold red]{var} (override failed: key not found)[/]')


def download_files(data: dict, target_folder: Path, offline_mode: bool, type: str):
	"""Download remote files to target folder, clone git repos, or symlink local files.

	Args:
		data (dict): Dict of name to url or local path prefixed with 'git+' for Git repos.
		target_folder (Path): Target folder for storing files or repos.
		type (str): Type of files to handle.
		offline_mode (bool): Offline mode.
	"""
	for name, url_or_path in data.items():
		target_path = download_file(url_or_path, target_folder, offline_mode, type, name=name)
		if target_path:
			data[name] = target_path


def download_file(url_or_path, target_folder: Path, offline_mode: bool, type: str, name: str = None):
	"""Download remote file to target folder, clone git repos, or symlink local files.

	Args:
		data (dict): Dict of name to url or local path prefixed with 'git+' for Git repos.
		target_folder (Path): Target folder for storing files or repos.
		offline_mode (bool): Offline mode.
		type (str): Type of files to handle.
		name (str, Optional): Name of object.

	Returns:
		path (Path): Path to downloaded file / folder.
	"""
	if url_or_path.startswith('git+'):
		# Clone Git repository
		git_url = url_or_path[4:]  # remove 'git+' prefix
		repo_name = git_url.split('/')[-1]
		if repo_name.endswith('.git'):
			repo_name = repo_name[:-4]
		target_path = target_folder / repo_name
		if not target_path.exists():
			console.print(f'[bold turquoise4]Cloning git {type} [bold magenta]{repo_name}[/] ...[/] ', end='')
			if offline_mode:
				console.print('[bold orange1]skipped [dim][offline[/].[/]')
				return
			try:
				call(['git', 'clone', git_url, str(target_path)], stderr=DEVNULL, stdout=DEVNULL)
				console.print('[bold green]ok.[/]')
			except Exception as e:
				console.print(f'[bold red]failed ({str(e)}).[/]')
		return target_path.resolve()
	elif Path(url_or_path).exists():
		# Create a symbolic link for a local file
		local_path = Path(url_or_path)
		target_path = target_folder / local_path.name
		if not name:
			name = url_or_path.split('/')[-1]
		if not CONFIG.security.allow_local_file_access:
			console.print(f'[bold red]Cannot reference local file {url_or_path}(disabled for security reasons)[/]')
			return
		if not target_path.exists():
			console.print(f'[bold turquoise4]Symlinking {type} [bold magenta]{name}[/] ...[/] ', end='')
			try:
				target_path.symlink_to(local_path)
				console.print('[bold green]ok.[/]')
			except Exception as e:
				console.print(f'[bold red]failed ({str(e)}).[/]')
		return target_path.resolve()
	else:
		# Download file from URL
		ext = url_or_path.split('.')[-1]
		if not name:
			name = url_or_path.split('/')[-1]
		filename = f'{name}.{ext}' if not name.endswith(ext) else name
		target_path = target_folder / filename
		if not target_path.exists():
			try:
				console.print(f'[bold turquoise4]Downloading {type} [bold magenta]{filename}[/] ...[/] ', end='')
				if offline_mode:
					console.print('[bold orange1]skipped [dim](offline)[/].[/]')
					return
				resp = requests.get(url_or_path, timeout=3)
				resp.raise_for_status()
				with open(target_path, 'wb') as f:
					f.write(resp.content)
				console.print('[bold green]ok.[/]')
			except requests.RequestException as e:
				console.print(f'[bold red]failed ({str(e)}).[/]')
				return
		return target_path.resolve()


# Load default_config
default_config = Config.parse(print_errors=False)

# Load user config
data_root = default_config.dirs.data
config_path = data_root / 'config.yml'
if not config_path.exists():
	if not data_root.exists():
		console.print(f'[bold turquoise4]Creating directory [bold magenta]{data_root}[/] ... [/]', end='')
		data_root.mkdir(parents=False)
		console.print('[bold green]ok.[/]')
	console.print(
		f'[bold turquoise4]Creating user conf [bold magenta]{config_path}[/]... [/]', end='')
	config_path.touch()
	console.print('[bold green]ok.[/]')
CONFIG = Config.parse(path=config_path)

# Fallback to default if invalid user config
if not CONFIG:
	console.print(f'[bold orange1]Invalid user config {config_path}. Falling back to default config.')
	CONFIG = default_config

# Create directories if they don't exist already
for name, dir in CONFIG.dirs.items():
	if not dir.exists():
		console.print(f'[bold turquoise4]Creating directory [bold magenta]{dir}[/] ... [/]', end='')
		dir.mkdir(parents=False)
		console.print('[bold green]ok.[/]')

# Download wordlists and payloads
download_files(CONFIG.wordlists.templates, CONFIG.dirs.wordlists, CONFIG.offline_mode, 'wordlist')
download_files(CONFIG.payloads.templates, CONFIG.dirs.payloads, CONFIG.offline_mode, 'payload')

# Print config
if CONFIG.debug.component == 'config':
	CONFIG.print()



import uuid

from secator.config import CONFIG
from secator.runners._base import Runner
from secator.runners.task import Task
from secator.utils import merge_opts
from secator.celery_utils import CeleryData
from secator.output_types import Info


class Workflow(Runner):

	default_exporters = CONFIG.workflows.exporters

	@classmethod
	def delay(cls, *args, **kwargs):
		from secator.celery import run_workflow
		return run_workflow.delay(args=args, kwargs=kwargs)

	def yielder(self):
		"""Run workflow.

		Yields:
			secator.output_types.OutputType: Secator output type.
		"""
		# Task opts
		run_opts = self.run_opts.copy()
		run_opts['hooks'] = self._hooks.get(Task, {})
		run_opts.pop('no_poll', False)

		# Build Celery workflow
		workflow = self.build_celery_workflow(
			run_opts=run_opts,
			results=self.results
		)
		self.celery_ids = list(self.celery_ids_map.keys())

		# Run Celery workflow and get results
		if self.sync:
			self.print_item = False
			results = workflow.apply().get()
		else:
			result = workflow()
			self.celery_ids.append(str(result.id))
			self.celery_result = result
			yield Info(
				message=f'Celery task created: {self.celery_result.id}',
				task_id=self.celery_result.id
			)
			if self.no_poll:
				return
			results = CeleryData.iter_results(
				self.celery_result,
				ids_map=self.celery_ids_map,
				description=True,
				print_remote_info=self.print_remote_info,
				print_remote_title=f'[bold gold3]{self.__class__.__name__.capitalize()}[/] [bold magenta]{self.name}[/] results'
			)

		# Get workflow results
		yield from results

	def build_celery_workflow(self, run_opts={}, results=[]):
		""""Build Celery workflow.

		Returns:
			tuple(celery.chain, List[str]): Celery task chain, Celery task ids.
		"""
		from celery import chain
		from secator.celery import forward_results
		sigs = self.get_tasks(
			self.config.tasks.toDict(),
			self.inputs,
			self.config.options,
			run_opts)
		sigs = [forward_results.si(results).set(queue='results')] + sigs + [forward_results.s().set(queue='results')]
		workflow = chain(*sigs)
		return workflow

	def get_tasks(self, config, inputs, workflow_opts, run_opts):
		"""Get tasks recursively as Celery chains / chords.

		Args:
			config (dict): Tasks config dict.
			inputs (list): Inputs.
			workflow_opts (dict): Workflow options.
			run_opts (dict): Run options.
			sync (bool): Synchronous mode (chain of tasks, no chords).

		Returns:
			tuple (List[celery.Signature], List[str]): Celery signatures, Celery task ids.
		"""
		from celery import chain, chord
		from secator.celery import forward_results
		sigs = []
		for task_name, task_opts in config.items():
			# Task opts can be None
			task_opts = task_opts or {}

			# If it's a group, process the sublevel tasks as a Celery chord.
			if task_name.startswith('_group'):
				tasks = self.get_tasks(
					task_opts,
					inputs,
					workflow_opts,
					run_opts
				)
				sig = chord((tasks), forward_results.s().set(queue='results'))
			elif task_name == '_chain':
				tasks = self.get_tasks(
					task_opts,
					inputs,
					workflow_opts,
					run_opts
				)
				sig = chain(*tasks)
			else:
				# Get task class
				task = Task.get_task_class(task_name)

				# Merge task options (order of priority with overrides)
				opts = merge_opts(workflow_opts, task_opts, run_opts)

				# Add task context and hooks to options
				opts['hooks'] = {task: self._hooks.get(Task, {})}
				opts['context'] = self.context.copy()
				opts['name'] = task_name
				opts['has_parent'] = True
				opts['skip_if_no_inputs'] = True

				# Create task signature
				task_id = str(uuid.uuid4())
				sig = task.s(inputs, **opts).set(queue=task.profile, task_id=task_id)
				self.add_subtask(task_id, task_name, task_opts.get('description', ''))
				self.output_types.extend(task.output_types)
			sigs.append(sig)
		return sigs



import time
from dataclasses import dataclass, field

from secator.output_types import OutputType
from secator.utils import rich_to_ansi, trim_string, rich_escape as _s


@dataclass
class Tag(OutputType):
	name: str
	match: str
	extra_data: dict = field(default_factory=dict, repr=True, compare=False)
	_source: str = field(default='', repr=True)
	_type: str = field(default='tag', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = ['match', 'name', 'extra_data']
	_sort_by = ('match', 'name')

	def __post_init__(self):
		super().__post_init__()

	def __str__(self) -> str:
		return self.match

	def __repr__(self) -> str:
		s = f'🏷️  [bold magenta]{self.name}[/]'
		s += f' found @ [bold]{_s(self.match)}[/]'
		ed = ''
		if self.extra_data:
			for k, v in self.extra_data.items():
				sep = ' '
				if not v:
					continue
				if isinstance(v, str):
					v = trim_string(v, max_length=1000)
					if len(v) > 1000:
						v = v.replace('\n', '\n' + sep)
						sep = '\n    '
				ed += f'\n    [dim red]{_s(k)}[/]:{sep}[dim yellow]{_s(v)}[/]'
		if ed:
			s += ed
		return rich_to_ansi(s)



import sys

from collections import OrderedDict

import rich_click as click
from rich_click.rich_click import _get_rich_console
from rich_click.rich_group import RichGroup

from secator.config import CONFIG
from secator.definitions import ADDONS_ENABLED, OPT_NOT_SUPPORTED
from secator.runners import Scan, Task, Workflow
from secator.utils import (deduplicate, expand_input, get_command_category)

RUNNER_OPTS = {
	'output': {'type': str, 'default': None, 'help': 'Output options (-o table,json,csv,gdrive)', 'short': 'o'},
	'workspace': {'type': str, 'default': 'default', 'help': 'Workspace', 'short': 'ws'},
	'print_json': {'is_flag': True, 'short': 'json', 'default': False, 'help': 'Print items as JSON lines'},
	'print_raw': {'is_flag': True, 'short': 'raw', 'default': False, 'help': 'Print items in raw format'},
	'print_stat': {'is_flag': True, 'short': 'stat', 'default': False, 'help': 'Print runtime statistics'},
	'print_format': {'default': '', 'short': 'fmt', 'help': 'Output formatting string'},
	'enable_profiler': {'is_flag': True, 'short': 'prof', 'default': False, 'help': 'Enable runner profiling'},
	'show': {'is_flag': True, 'default': False, 'help': 'Show command that will be run (tasks only)'},
	'no_process': {'is_flag': True, 'default': False, 'help': 'Disable secator processing'},
	# 'filter': {'default': '', 'short': 'f', 'help': 'Results filter', 'short': 'of'}, # TODO add this
	'quiet': {'is_flag': True, 'default': False, 'help': 'Enable quiet mode'},
}

RUNNER_GLOBAL_OPTS = {
	'sync': {'is_flag': True, 'help': 'Run tasks synchronously (automatic if no worker is alive)'},
	'worker': {'is_flag': True, 'default': False, 'help': 'Run tasks in worker'},
	'no_poll': {'is_flag': True, 'default': False, 'help': 'Do not live poll for tasks results when running in worker'},
	'proxy': {'type': str, 'help': 'HTTP proxy'},
	'driver': {'type': str, 'help': 'Export real-time results. E.g: "mongodb"'}
	# 'debug': {'type': int, 'default': 0, 'help': 'Debug mode'},
}

DEFAULT_CLI_OPTIONS = list(RUNNER_OPTS.keys()) + list(RUNNER_GLOBAL_OPTS.keys())


class OrderedGroup(RichGroup):
	def __init__(self, name=None, commands=None, **attrs):
		super(OrderedGroup, self).__init__(name, commands, **attrs)
		self.commands = commands or OrderedDict()

	def command(self, *args, **kwargs):
		"""Behaves the same as `click.Group.command()` but supports aliases.
		"""
		def decorator(f):
			aliases = kwargs.pop("aliases", None)
			if aliases:
				max_width = _get_rich_console().width
				aliases_str = ', '.join(f'[bold cyan]{alias}[/]' for alias in aliases)
				padding = max_width // 4

				name = kwargs.pop("name", None)
				if not name:
					raise click.UsageError("`name` command argument is required when using aliases.")

				f.__doc__ = f.__doc__ or '\0'.ljust(padding+1)
				f.__doc__ = f'{f.__doc__:<{padding}}[dim](aliases)[/] {aliases_str}'
				base_command = super(OrderedGroup, self).command(
					name, *args, **kwargs
				)(f)
				for alias in aliases:
					cmd = super(OrderedGroup, self).command(alias, *args, hidden=True, **kwargs)(f)
					cmd.help = f"Alias for '{name}'.\n\n{cmd.help}"
					cmd.params = base_command.params

			else:
				cmd = super(OrderedGroup, self).command(*args, **kwargs)(f)

			return cmd
		return decorator

	def group(self, *args, **kwargs):
		"""Behaves the same as `click.Group.group()` but supports aliases.
		"""
		def decorator(f):
			aliases = kwargs.pop('aliases', [])
			aliased_group = []
			if aliases:
				max_width = _get_rich_console().width
				aliases_str = ', '.join(f'[bold cyan]{alias}[/]' for alias in aliases)
				padding = max_width // 4
				f.__doc__ = f.__doc__ or '\0'.ljust(padding+1)
				f.__doc__ = f'{f.__doc__:<{padding}}[dim](aliases)[/] {aliases_str}'
				for alias in aliases:
					grp = super(OrderedGroup, self).group(
						alias, *args, hidden=True, **kwargs)(f)
					aliased_group.append(grp)

			# create the main group
			grp = super(OrderedGroup, self).group(*args, **kwargs)(f)
			grp.aliases = aliases

			# for all of the aliased groups, share the main group commands
			for aliased in aliased_group:
				aliased.commands = grp.commands

			return grp
		return decorator

	def list_commands(self, ctx):
		return self.commands


def get_command_options(config):
	"""Get unified list of command options from a list of secator tasks classes and optionally a Runner config.

	Args:
		config (TemplateLoader): Current runner config.

	Returns:
		list: List of deduplicated options.
	"""
	from secator.utils import debug
	opt_cache = []
	all_opts = OrderedDict({})
	tasks = config.flat_tasks
	tasks_cls = set([c['class'] for c in tasks.values()])

	# Loop through tasks and set options
	for cls in tasks_cls:
		opts = OrderedDict(RUNNER_GLOBAL_OPTS, **RUNNER_OPTS, **cls.meta_opts, **cls.opts)

		# Find opts defined in config corresponding to this task class
		# TODO: rework this as this ignores subsequent tasks of the same task class
		task_config_opts = {}
		if config.type != 'task':
			for k, v in tasks.items():
				if v['class'] == cls:
					task_config_opts = v['opts']

		# Loop through options
		for opt, opt_conf in opts.items():

			# Get opt key map if any
			opt_key_map = getattr(cls, 'opt_key_map', {})

			# Opt is not supported by this task
			if opt not in opt_key_map\
				and opt not in cls.opts\
				and opt not in RUNNER_OPTS\
				and opt not in RUNNER_GLOBAL_OPTS:
				continue

			# Opt is defined as unsupported
			if opt_key_map.get(opt) == OPT_NOT_SUPPORTED:
				continue

			# Get opt prefix
			prefix = None
			if opt in cls.opts:
				prefix = cls.__name__
			elif opt in cls.meta_opts:
				# TODO: Add options categories
				# category = get_command_category(cls)
				# prefix = category
				prefix = 'Meta'
			elif opt in RUNNER_OPTS:
				prefix = 'Output'
			elif opt in RUNNER_GLOBAL_OPTS:
				prefix = 'Execution'

			# Get opt conf
			conf = opt_conf.copy()
			conf['show_default'] = True
			conf['prefix'] = prefix
			opt_default = conf.get('default', None)
			opt_is_flag = conf.get('is_flag', False)
			opt_value_in_config = task_config_opts.get(opt)

			# Check if opt already defined in config
			if opt_value_in_config:
				if conf.get('required', False):
					debug('OPT (skipped: opt is required and defined in config)', obj={'opt': opt}, sub=f'cli.{config.name}', verbose=True)  # noqa: E501
					continue
				mapped_value = cls.opt_value_map.get(opt)
				if callable(mapped_value):
					opt_value_in_config = mapped_value(opt_value_in_config)
				elif mapped_value:
					opt_value_in_config = mapped_value
				if opt_value_in_config != opt_default:
					if opt in opt_cache:
						continue
					if opt_is_flag:
						conf['reverse'] = True
						conf['default'] = not conf['default']
					# print(f'{opt}: change default to {opt_value_in_config}')
					conf['default'] = opt_value_in_config

			# If opt is a flag but the default is True, add opposite flag
			if opt_is_flag and opt_default is True:
				conf['reverse'] = True

			# Check if opt already processed before
			if opt in opt_cache:
				# debug('OPT (skipped: opt is already in opt cache)', obj={'opt': opt}, sub=f'cli.{config.name}', verbose=True)
				continue

			# Build help
			opt_cache.append(opt)
			opt = opt.replace('_', '-')
			all_opts[opt] = conf

			# Debug
			debug_conf = OrderedDict({'opt': opt, 'config_val': opt_value_in_config or 'N/A', **conf.copy()})
			debug('OPT', obj=debug_conf, sub=f'cli.{config.name}', verbose=True)

	return all_opts


def decorate_command_options(opts):
	"""Add click.option decorator to decorate click command.

	Args:
		opts (dict): Dict of command options.

	Returns:
		function: Decorator.
	"""
	def decorator(f):
		reversed_opts = OrderedDict(list(opts.items())[::-1])
		for opt_name, opt_conf in reversed_opts.items():
			conf = opt_conf.copy()
			short_opt = conf.pop('short', None)
			conf.pop('internal', None)
			conf.pop('prefix', None)
			conf.pop('shlex', None)
			conf.pop('meta', None)
			conf.pop('supported', None)
			conf.pop('process', None)
			reverse = conf.pop('reverse', False)
			long = f'--{opt_name}'
			short = f'-{short_opt}' if short_opt else f'-{opt_name}'
			if reverse:
				long += f'/--no-{opt_name}'
				short += f'/-n{short_opt}' if short else f'/-n{opt_name}'
			f = click.option(long, short, **conf)(f)
		return f
	return decorator


def task():
	def decorator(cls):
		cls.__task__ = True
		return cls
	return decorator


def generate_cli_subcommand(cli_endpoint, func, **opts):
	return cli_endpoint.command(**opts)(func)


def register_runner(cli_endpoint, config):
	name = config.name
	input_required = True
	input_type = 'targets'
	command_opts = {
		'no_args_is_help': True,
		'context_settings': {
			'ignore_unknown_options': False,
			'allow_extra_args': False
		}
	}

	if cli_endpoint.name == 'scan':
		runner_cls = Scan
		short_help = config.description or ''
		short_help += f' [dim]alias: {config.alias}' if config.alias else ''
		command_opts.update({
			'name': name,
			'short_help': short_help
		})

	elif cli_endpoint.name == 'workflow':
		runner_cls = Workflow
		short_help = config.description or ''
		short_help = f'{short_help:<55} [dim](alias)[/][bold cyan] {config.alias}' if config.alias else ''
		command_opts.update({
			'name': name,
			'short_help': short_help
		})

	elif cli_endpoint.name == 'task':
		runner_cls = Task
		input_required = False  # allow targets from stdin
		task_cls = Task.get_task_class(config.name)
		task_category = get_command_category(task_cls)
		input_type = task_cls.input_type or 'targets'
		short_help = f'[magenta]{task_category:<15}[/]{task_cls.__doc__}'
		command_opts.update({
			'name': name,
			'short_help': short_help,
			'no_args_is_help': False
		})

	else:
		raise ValueError(f"Unrecognized runner endpoint name {cli_endpoint.name}")
	options = get_command_options(config)

	# TODO: maybe allow this in the future
	# def get_unknown_opts(ctx):
	# 	return {
	# 		(ctx.args[i][2:]
	# 		if str(ctx.args[i]).startswith("--") \
	# 		else ctx.args[i][1:]): ctx.args[i+1]
	# 		for i in range(0, len(ctx.args), 2)
	# 	}

	@click.argument(input_type, required=input_required)
	@decorate_command_options(options)
	@click.pass_context
	def func(ctx, **opts):
		sync = opts['sync']
		worker = opts.pop('worker')
		ws = opts.pop('workspace')
		driver = opts.pop('driver', '')
		show = opts['show']
		context = {'workspace_name': ws}

		# Remove options whose values are default values
		for k, v in options.items():
			opt_name = k.replace('-', '_')
			if opt_name in opts and opts[opt_name] == v.get('default', None):
				del opts[opt_name]

		# TODO: maybe allow this in the future
		# unknown_opts = get_unknown_opts(ctx)
		# opts.update(unknown_opts)

		# Expand input
		inputs = opts.pop(input_type)
		inputs = expand_input(inputs, ctx)

		# Build hooks from driver name
		hooks = []
		drivers = driver.split(',') if driver else []
		console = _get_rich_console()
		supported_drivers = ['mongodb', 'gcs']
		for driver in drivers:
			if driver in supported_drivers:
				if not ADDONS_ENABLED[driver]:
					console.print(f'[bold red]Missing "{driver}" addon: please run `secator install addons {driver}`[/].')
					sys.exit(1)
				from secator.utils import import_dynamic
				driver_hooks = import_dynamic(f'secator.hooks.{driver}', 'HOOKS')
				if driver_hooks is None:
					console.print(f'[bold red]Missing "secator.hooks.{driver}.HOOKS".[/]')
					sys.exit(1)
				hooks.append(driver_hooks)
			else:
				supported_drivers_str = ', '.join([f'[bold green]{_}[/]' for _ in supported_drivers])
				console.print(f'[bold red]Driver "{driver}" is not supported.[/]')
				console.print(f'Supported drivers: {supported_drivers_str}')
				sys.exit(1)

		from secator.utils import deep_merge_dicts
		hooks = deep_merge_dicts(*hooks)

		# Enable sync or not
		if sync or show:
			sync = True
		else:
			from secator.celery import is_celery_worker_alive
			worker_alive = is_celery_worker_alive()
			if not worker_alive and not worker:
				sync = True
			else:
				sync = False
				broker_protocol = CONFIG.celery.broker_url.split('://')[0]
				backend_protocol = CONFIG.celery.result_backend.split('://')[0]
				if CONFIG.celery.broker_url:
					if (broker_protocol == 'redis' or backend_protocol == 'redis') and not ADDONS_ENABLED['redis']:
						_get_rich_console().print('[bold red]Missing `redis` addon: please run `secator install addons redis`[/].')
						sys.exit(1)

		from secator.utils import debug
		debug('Run options', obj=opts, sub='cli')

		# Set run options
		opts.update({
			'print_cmd': True,
			'print_item': True,
			'print_line': True,
			'print_progress': True,
			'print_remote_info': not sync,
			'piped_input': ctx.obj['piped_input'],
			'piped_output': ctx.obj['piped_output'],
			'caller': 'cli',
			'sync': sync,
		})

		# Start runner
		runner = runner_cls(config, inputs, run_opts=opts, hooks=hooks, context=context)
		runner.run()

	generate_cli_subcommand(cli_endpoint, func, **command_opts)
	generate_rich_click_opt_groups(cli_endpoint, name, input_type, options)


def generate_rich_click_opt_groups(cli_endpoint, name, input_type, options):
	sortorder = {
		'Execution': 0,
		'Output': 1,
		'Meta': 2,
	}
	prefixes = deduplicate([opt['prefix'] for opt in options.values()])
	prefixes = sorted(prefixes, key=lambda x: sortorder.get(x, 3))
	opt_group = [
		{
			'name': 'Targets',
			'options': [input_type],
		},
	]
	for prefix in prefixes:
		prefix_opts = [
			opt for opt, conf in options.items()
			if conf['prefix'] == prefix
		]
		opt_names = [f'--{opt_name}' for opt_name in prefix_opts]
		if prefix == 'Execution':
			opt_names.append('--help')
		opt_group.append({
			'name': prefix + ' options',
			'options': opt_names
		})
	aliases = [cli_endpoint.name, *cli_endpoint.aliases]
	for alias in aliases:
		endpoint_name = f'secator {alias} {name}'
		click.rich_click.OPTION_GROUPS[endpoint_name] = opt_group



__all__ = [
	'Error',
	'OutputType',
	'Info',
	'Ip',
	'Port',
	'Progress',
	'Record',
	'Stat',
	'Subdomain',
	'Url',
	'UserAccount',
	'Vulnerability',
	'Warning',
]
from secator.output_types._base import OutputType
from secator.output_types.progress import Progress
from secator.output_types.ip import Ip
from secator.output_types.exploit import Exploit
from secator.output_types.port import Port
from secator.output_types.subdomain import Subdomain
from secator.output_types.tag import Tag
from secator.output_types.target import Target
from secator.output_types.url import Url
from secator.output_types.user_account import UserAccount
from secator.output_types.vulnerability import Vulnerability
from secator.output_types.record import Record
from secator.output_types.info import Info
from secator.output_types.warning import Warning
from secator.output_types.error import Error
from secator.output_types.stat import Stat

EXECUTION_TYPES = [
	Target, Progress, Info, Warning, Error
]
STAT_TYPES = [
	Stat
]
FINDING_TYPES = [
	Subdomain, Ip, Port, Url, Tag, Exploit, UserAccount, Vulnerability
]
OUTPUT_TYPES = FINDING_TYPES + EXECUTION_TYPES + STAT_TYPES



import operator

from secator.config import CONFIG
from secator.output_types import FINDING_TYPES, OutputType
from secator.utils import merge_opts, get_file_timestamp, traceback_as_string
from secator.rich import console
from secator.runners._helpers import extract_from_results

import concurrent.futures
from threading import Lock


def remove_duplicates(objects):
	unique_objects = []
	lock = Lock()

	def add_if_unique(obj):
		nonlocal unique_objects
		with lock:
			# Perform linear search to check for duplicates
			if all(obj != existing_obj for existing_obj in unique_objects):
				unique_objects.append(obj)

	with concurrent.futures.ThreadPoolExecutor(max_workers=100) as executor:
		# Execute the function concurrently for each object
		executor.map(add_if_unique, objects)

	return unique_objects


# TODO: initialize from data, not from runner
class Report:
	"""Report class.

	Args:
		runner (secator.runners.Runner): Runner instance.
		title (str): Report title.
		exporters (list): List of exporter classes.
	"""
	def __init__(self, runner, title=None, exporters=[]):
		self.title = title or f'{runner.__class__.__name__.lower()}_{runner.config.name}'
		self.runner = runner
		self.timestamp = get_file_timestamp()
		self.exporters = exporters
		self.workspace_name = runner.workspace_name
		self.output_folder = runner.reports_folder

	def send(self):
		for report_cls in self.exporters:
			try:
				report_cls(self).send()
			except Exception as e:
				console.print(
					f'[bold red]Could not create exporter {report_cls.__name__} for {self.__class__.__name__}: '
					f'{str(e)}[/]\n[dim]{traceback_as_string(e)}[/]',
				)

	def build(self, extractors=[], dedupe=False):
		# Trim options
		from secator.decorators import DEFAULT_CLI_OPTIONS
		opts = merge_opts(self.runner.config.options, self.runner.run_opts)
		opts = {
			k: v for k, v in opts.items()
			if k not in DEFAULT_CLI_OPTIONS and k not in self.runner.print_opts
			and v is not None
		}
		runner_fields = {
			'name',
			'status',
			'targets',
			'start_time',
			'end_time',
			'elapsed',
			'elapsed_human',
			'run_opts',
			'results_count'
		}

		# Prepare report structure
		data = {
			'info': {k: v for k, v in self.runner.toDict().items() if k in runner_fields},
			'results': {}
		}
		if 'results' in data['info']:
			del data['info']['results']
		data['info']['title'] = self.title
		data['info']['errors'] = self.runner.errors

		# Fill report
		for output_type in FINDING_TYPES:
			output_name = output_type.get_name()
			sort_by, _ = get_table_fields(output_type)
			items = [
				item for item in self.runner.results
				if isinstance(item, OutputType) and item._type == output_name
			]
			if items:
				if sort_by and all(sort_by):
					items = sorted(items, key=operator.attrgetter(*sort_by))
				if dedupe and CONFIG.runners.remove_duplicates:
					items = remove_duplicates(items)
					# items = [item for item in items if not item._duplicate and item not in dedupe_from]
				for extractor in extractors:
					items = extract_from_results(items, extractors=[extractor])
				data['results'][output_name] = items

		# Save data
		self.data = data

	def is_empty(self):
		return all(not items for items in self.data['results'].values())


def get_table_fields(output_type):
	"""Get output fields and sort fields based on output type.

	Args:
		output_type (str): Output type.

	Returns:
		tuple: Tuple of sort_by (tuple), output_fields (list).
	"""
	sort_by = ()
	output_fields = []
	if output_type in FINDING_TYPES:
		sort_by = output_type._sort_by
		output_fields = output_type._table_fields
	return sort_by, output_fields



import os
import signal
import threading
from pathlib import Path

from celery import signals

from secator.config import CONFIG
from secator.output_types import Info
from secator.rich import console

IDLE_TIMEOUT = CONFIG.celery.worker_kill_after_idle_seconds

# File-based state management system
STATE_DIR = Path("/tmp/celery_state")
STATE_DIR.mkdir(exist_ok=True, parents=True)


def get_lock_file_path():
    worker_name = os.environ.get("WORKER_NAME", f"unknown_{os.getpid()}")
    return Path(f"/tmp/celery_worker_{worker_name}.lock")


def set_task_running(task_id):
    """Mark that a task is running in current worker"""
    with open(get_lock_file_path(), "w") as f:
        f.write(task_id)


def clear_task_running():
    """Clear the task running state"""
    lock_file = get_lock_file_path()
    if lock_file.exists():
        lock_file.unlink()


def is_task_running():
    """Check if a task is currently running"""
    return get_lock_file_path().exists()


def kill_worker(parent=False):
    """Kill current worker using its pid by sending a SIGTERM to Celery master process."""
    worker_name = os.environ.get('WORKER_NAME', 'unknown')

    # Check if a task is running via the lock file
    if not is_task_running():
        pid = os.getppid() if parent else os.getpid()
        console.print(Info(message=f'Sending SIGTERM to worker {worker_name} with pid {pid}'))
        os.kill(pid, signal.SIGTERM)
    else:
        console.print(Info(message=f'Cancelling worker shutdown of {worker_name} since a task is running'))


def setup_idle_timer(timeout):
    """Setup a timer to kill the worker after being idle"""
    if timeout == -1:
        return

    console.print(Info(message=f'Starting inactivity timer for {timeout} seconds ...'))
    timer = threading.Timer(timeout, kill_worker)
    timer.daemon = True  # Make sure timer is killed when worker exits
    timer.start()


def maybe_override_logging():
    def decorator(func):
        if CONFIG.celery.override_default_logging:
            return signals.setup_logging.connect(func)
        else:
            return func
    return decorator


@maybe_override_logging()
def setup_logging(*args, **kwargs):
    """Override celery's logging setup to prevent it from altering our settings.
    github.com/celery/celery/issues/1867
    """
    pass


def capture_worker_name(sender, instance, **kwargs):
    os.environ["WORKER_NAME"] = '{0}'.format(sender)


def worker_init_handler(**kwargs):
    if IDLE_TIMEOUT != -1:
        setup_idle_timer(IDLE_TIMEOUT)


def task_prerun_handler(task_id, **kwargs):
    # Mark that a task is running
    set_task_running(task_id)


def task_postrun_handler(**kwargs):
    # Mark that no task is running
    clear_task_running()

    # Get sender name from kwargs
    sender_name = kwargs['sender'].name

    if CONFIG.celery.worker_kill_after_task and sender_name.startswith('secator.'):
        worker_name = os.environ.get('WORKER_NAME', 'unknown')
        console.print(Info(message=f'Shutdown worker {worker_name} since config celery.worker_kill_after_task is set.'))
        kill_worker(parent=True)
        return

    # Set up a new idle timer
    if IDLE_TIMEOUT != -1:
        console.print(Info(message=f'Reset inactivity timer to {IDLE_TIMEOUT} seconds'))
        setup_idle_timer(IDLE_TIMEOUT)


def task_revoked_handler(request=None, **kwargs):
    """Handle revoked tasks by clearing the task running state"""
    console.print(Info(message='Task was revoked, clearing running state'))
    clear_task_running()

    # Set up a new idle timer
    if IDLE_TIMEOUT != -1:
        console.print(Info(message=f'Reset inactivity timer to {IDLE_TIMEOUT} seconds after task revocation'))
        setup_idle_timer(IDLE_TIMEOUT)


def worker_shutdown_handler(**kwargs):
    """Cleanup lock files when worker shuts down"""
    lock_file = get_lock_file_path()
    if lock_file.exists():
        lock_file.unlink()


def setup_handlers():
    signals.celeryd_after_setup.connect(capture_worker_name)
    signals.setup_logging.connect(setup_logging)
    signals.task_prerun.connect(task_prerun_handler)
    signals.task_postrun.connect(task_postrun_handler)
    signals.task_revoked.connect(task_revoked_handler)
    signals.worker_ready.connect(worker_init_handler)
    signals.worker_shutdown.connect(worker_shutdown_handler)



import json
import logging
import os
import sys
import uuid
from datetime import datetime
from pathlib import Path
from time import time

import humanize

from secator.definitions import ADDONS_ENABLED
from secator.celery_utils import CeleryData
from secator.config import CONFIG
from secator.output_types import FINDING_TYPES, OutputType, Progress, Info, Warning, Error, Target
from secator.report import Report
from secator.rich import console, console_stdout
from secator.runners._helpers import (get_task_folder_id, process_extractor, run_extractors)
from secator.utils import (debug, import_dynamic, merge_opts, rich_to_ansi, should_update)

logger = logging.getLogger(__name__)

HOOKS = [
	'before_init',
	'on_init',
	'on_start',
	'on_end',
	'on_item_pre_convert',
	'on_item',
	'on_duplicate',
	'on_interval',
]

VALIDATORS = [
	'validate_input',
	'validate_item'
]


class Runner:
	"""Runner class.

	Args:
		config (secator.config.TemplateLoader): Runner config.
		inputs (List[str]): List of inputs to run task on.
		results (List[OutputType]): List of results to re-use.
		run_opts (dict[str]): Run options.
		hooks (dict[str, List[Callable]]): User hooks to register.
		validators (dict): User validators to register.
		context (dict): Runner context.

	Yields:
		OutputType: Output types.
	"""

	# Input field (mostly for tests and CLI)
	input_type = None

	# Output types
	output_types = []

	# Default exporters
	default_exporters = []

	# Run hooks
	enable_hooks = True

	# Reports folder
	reports_folder = None

	def __init__(self, config, inputs=[], results=[], run_opts={}, hooks={}, validators={}, context={}):
		self.uuids = []
		self.results = []
		self.output = ''

		# Runner config
		self.config = config
		self.name = run_opts.get('name', config.name)
		self.description = run_opts.get('description', config.description)
		self.workspace_name = context.get('workspace_name', 'default')
		self.run_opts = run_opts.copy()
		self.sync = run_opts.get('sync', True)
		self.done = False
		self.start_time = datetime.fromtimestamp(time())
		self.last_updated_db = None
		self.last_updated_celery = None
		self.last_updated_progress = None
		self.end_time = None
		self._hooks = hooks
		self.progress = 0
		self.context = context
		self.delay = run_opts.get('delay', False)
		self.celery_result = None
		self.celery_ids = []
		self.celery_ids_map = {}
		self.caller = self.run_opts.get('caller', None)
		self.threads = []
		self.no_poll = self.run_opts.get('no_poll', False)
		self.quiet = self.run_opts.get('quiet', False)

		# Runner process options
		self.no_process = self.run_opts.get('no_process', False)
		self.piped_input = self.run_opts.get('piped_input', False)
		self.piped_output = self.run_opts.get('piped_output', False)
		self.enable_duplicate_check = self.run_opts.get('enable_duplicate_check', True)

		# Runner print opts
		self.print_item = self.run_opts.get('print_item', False)
		self.print_line = self.run_opts.get('print_line', False) and not self.quiet
		self.print_remote_info = self.run_opts.get('print_remote_info', False) and not self.piped_input and not self.piped_output  # noqa: E501
		self.print_json = self.run_opts.get('print_json', False)
		self.print_raw = self.run_opts.get('print_raw', False) or self.piped_output
		self.print_fmt = self.run_opts.get('fmt', '')
		self.print_progress = self.run_opts.get('print_progress', False) and not self.quiet and not self.print_raw
		self.print_target = self.run_opts.get('print_target', False) and not self.quiet and not self.print_raw
		self.print_stat = self.run_opts.get('print_stat', False) and not self.quiet and not self.print_raw
		self.raise_on_error = self.run_opts.get('raise_on_error', False)
		self.print_opts = {k: v for k, v in self.__dict__.items() if k.startswith('print_') if v}

		# Determine inputs
		inputs = [inputs] if not isinstance(inputs, list) else inputs
		if results:
			inputs, run_opts, errors = run_extractors(results, run_opts, inputs)
			for error in errors:
				self.add_result(error, print=True)
		self.inputs = inputs

		# Debug
		self.debug('Inputs', obj=self.inputs, sub='init')
		self.debug('Run opts', obj={k: v for k, v in self.run_opts.items() if v is not None}, sub='init')
		self.debug('Print opts', obj={k: v for k, v in self.print_opts.items() if v is not None}, sub='init')

		# Determine exporters
		exporters_str = self.run_opts.get('output') or self.default_exporters
		self.exporters = self.resolve_exporters(exporters_str)

		# Determine report folder
		default_reports_folder_base = f'{CONFIG.dirs.reports}/{self.workspace_name}/{self.config.type}s'
		_id = get_task_folder_id(default_reports_folder_base)
		self.reports_folder = f'{default_reports_folder_base}/{_id}'

		# Make reports folders
		os.makedirs(self.reports_folder, exist_ok=True)
		os.makedirs(f'{self.reports_folder}/.inputs', exist_ok=True)
		os.makedirs(f'{self.reports_folder}/.outputs', exist_ok=True)

		# Profiler
		self.enable_profiler = self.run_opts.get('enable_profiler', False) and ADDONS_ENABLED['trace']
		if self.enable_profiler:
			from pyinstrument import Profiler
			self.profiler = Profiler(async_mode=False, interval=0.0001)
			try:
				self.profiler.start()
			except RuntimeError:
				self.enable_profiler = False
				pass

		# Hooks
		self.hooks = {name: [] for name in HOOKS + getattr(self, 'hooks', [])}
		self.register_hooks(hooks)

		# Validators
		self.validators = {name: [] for name in VALIDATORS + getattr(self, 'validators', [])}
		self.register_validators(validators)

		# Chunks
		self.has_parent = self.run_opts.get('has_parent', False)
		self.has_children = self.run_opts.get('has_children', False)
		self.chunk = self.run_opts.get('chunk', None)
		self.chunk_count = self.run_opts.get('chunk_count', None)
		self.unique_name = self.name.replace('/', '_')
		self.unique_name = f'{self.unique_name}_{self.chunk}' if self.chunk else self.unique_name

		# Process prior results
		for result in results:
			list(self._process_item(result, print=False, output=False))

		# Input post-process
		self.run_hooks('before_init')

		# Check if input is valid
		self.inputs_valid = self.run_validators('validate_input', self.inputs)

		# Run hooks
		self.run_hooks('on_init')

	@property
	def elapsed(self):
		if self.done:
			return self.end_time - self.start_time
		return datetime.fromtimestamp(time()) - self.start_time

	@property
	def elapsed_human(self):
		return humanize.naturaldelta(self.elapsed)

	@property
	def targets(self):
		return [r for r in self.results if isinstance(r, Target)]

	@property
	def infos(self):
		return [r for r in self.results if isinstance(r, Info)]

	@property
	def warnings(self):
		return [r for r in self.results if isinstance(r, Warning)]

	@property
	def errors(self):
		return [r for r in self.results if isinstance(r, Error)]

	@property
	def self_results(self):
		return [r for r in self.results if r._source.startswith(self.unique_name)]

	@property
	def findings(self):
		return [r for r in self.results if isinstance(r, tuple(FINDING_TYPES))]

	@property
	def findings_count(self):
		return len(self.findings)

	@property
	def self_findings(self):
		return [r for r in self.results if isinstance(r, tuple(FINDING_TYPES)) if r._source.startswith(self.unique_name)]

	@property
	def self_errors(self):
		if self.config.type == 'task':
			return [r for r in self.results if isinstance(r, Error) and r._source.startswith(self.unique_name)]
		return [r for r in self.results if isinstance(r, Error)]

	@property
	def self_findings_count(self):
		return len(self.self_findings)

	@property
	def status(self):
		if not self.done:
			return 'RUNNING'
		return 'FAILURE' if len(self.self_errors) > 0 else 'SUCCESS'

	@property
	def celery_state(self):
		return {
			'name': self.config.name,
			'full_name': self.unique_name,
			'state': self.status,
			'progress': self.progress,
			'results': self.self_results,
			'chunk': self.chunk,
			'chunk_count': self.chunk_count,
			'chunk_info': f'{self.chunk}/{self.chunk_count}' if self.chunk and self.chunk_count else '',
			'celery_id': self.context['celery_id'],
			'count': self.self_findings_count,
			'descr': self.config.description or '',
		}

	def run(self):
		"""Run method.

		Returns:
			List[OutputType]: List of runner results.
		"""
		return list(self.__iter__())

	def __iter__(self):
		"""Process results from derived runner class in real-time and yield results.

		Yields:
			OutputType: runner result.
		"""
		try:
			self.log_start()
			self.run_hooks('on_start')

			# If any errors happened during valid ation, exit
			if self.errors:
				yield from self.errors
				self.log_results()
				self.run_hooks('on_end')
				return

			# Choose yielder
			yielder = self.yielder_celery if self.celery_result else self.yielder

			# Loop and process items
			for item in yielder():
				yield from self._process_item(item)
				self.run_hooks('on_interval')

			# Wait for threads to finish
			yield from self.join_threads()

		except BaseException as e:
			self.debug(f'encountered exception {type(e).__name__}. Stopping remote tasks.', sub='error')
			error = Error.from_exception(e)
			error._source = self.unique_name
			error._uuid = str(uuid.uuid4())
			self.add_result(error, print=True)
			self.stop_celery_tasks()
			yield from self.join_threads()
			yield error

		# Mark duplicates and filter results
		if not self.no_process:
			self.mark_duplicates()
			self.results = self.filter_results()

		# Finalize run
		self.log_results()
		self.run_hooks('on_end')

	def join_threads(self):
		"""Wait for all running threads to complete."""
		if not self.threads:
			return
		self.debug(f'waiting for {len(self.threads)} threads to complete')
		for thread in self.threads:
			error = thread.join()
			if error:
				error._source = self.unique_name
				error._uuid = str(uuid.uuid4())
				self.add_result(error, print=True)
				yield error

	def add_result(self, item, print=False):
		"""Add item to runner results.

		Args:
			item (OutputType): Item.
			print (bool): Whether to print it or not.
		"""
		self.uuids.append(item._uuid)
		self.results.append(item)
		self.output += repr(item) + '\n'
		if print:
			self._print_item(item)

	def add_subtask(self, task_id, task_name, task_description):
		"""Add a Celery subtask to the current runner for tracking purposes.

		Args:
			task_id (str): Celery task id.
			task_name (str): Task name.
			task_description (str): Task description.
		"""
		self.celery_ids.append(task_id)
		self.celery_ids_map[task_id] = {
			'id': task_id,
			'name': task_name,
			'full_name': task_name,
			'descr': task_description,
			'state': 'PENDING',
			'count': 0,
			'progress': 0
		}

	def _print_item(self, item, force=False):
		"""Print an item and add it to the runner's output.

		Args:
			item (str | OutputType): Secator output type to print.
			force (bool): Whether to force-print it.
		"""
		item_str = str(item)

		# Item is an output type
		if isinstance(item, OutputType):
			self.debug(item, lazy=lambda x: repr(x), sub='item', allow_no_process=False, verbose=True)
			_type = item._type
			print_this_type = getattr(self, f'print_{_type}', True)
			if not print_this_type:
				return

			if self.print_item or force:
				item_out = sys.stdout

				# JSON lines output
				if self.print_json:
					self._print(item, out=sys.stdout)
					item_out = None  # suppress item repr output to sdout

				# Raw output
				elif self.print_raw:
					item_out = sys.stderr if self.piped_output else None

					# Format raw output with custom item fields
					if self.print_fmt:
						try:
							item_str = item.format(**self.print_fmt)
						except KeyError:
							item_str = ''

					# raw output is used to pipe, we should only pipe the first output type of a Runner.
					if not isinstance(item, self.output_types[0]):
						item_str = ''

					if item_str:
						self._print(item_str, out=sys.stdout)

				# Repr output
				if item_out:
					item_repr = repr(item)
					if isinstance(item, OutputType) and self.print_remote_info:
						item_repr += rich_to_ansi(rf' \[[dim]{item._source}[/]]')
					self._print(item_repr, out=item_out)

		# Item is a line
		elif isinstance(item, str):
			self.debug(item, sub='line', allow_no_process=False, verbose=True)
			if self.print_line or force:
				self._print(item, out=sys.stderr, end='\n', rich=False)

	def debug(self, *args, **kwargs):
		"""Print debug with runner class name, only if self.no_process is True.

		Args:
			args (list): List of debug args.
			kwargs (dict): Dict of debug kwargs.
		"""
		allow_no_process = kwargs.pop('allow_no_process', True)
		if self.no_process and not allow_no_process:
			return
		sub = kwargs.get('sub')
		new_sub = f'runner.{self.__class__.__name__}'
		if sub:
			new_sub += f'.{sub}'
		kwargs['sub'] = new_sub
		debug(*args, **kwargs)

	def mark_duplicates(self):
		"""Check for duplicates and mark items as duplicates."""
		if not self.enable_duplicate_check:
			return
		self.debug('running duplicate check', id=self.config.name, sub='duplicates')
		# dupe_count = 0
		import concurrent.futures
		executor = concurrent.futures.ThreadPoolExecutor(max_workers=100)
		for item in self.results.copy():
			executor.submit(self.check_duplicate, item)
		executor.shutdown(wait=True)
		# duplicates = [repr(i) for i in self.results if i._duplicate]
		# if duplicates:
		# 	duplicates_str = '\n\t'.join(duplicates)
		# 	self.debug(f'Duplicates ({dupe_count}):\n\t{duplicates_str}', sub='duplicates', verbose=True)
		# self.debug(f'duplicate check completed: {dupe_count} found', id=self.config.name, sub='duplicates')

	def check_duplicate(self, item):
		"""Check if an item is a duplicate in the list of results and mark it like so.

		Args:
			item (OutputType): Secator output type.
		"""
		self.debug('running duplicate check for item', obj=item.toDict(), obj_breaklines=True, sub='duplicates', verbose=True)
		others = [f for f in self.results if f == item and f._uuid != item._uuid]
		if others:
			main = max(item, *others)
			dupes = [f for f in others if f._uuid != main._uuid]
			main._duplicate = False
			main._related.extend([dupe._uuid for dupe in dupes])
			main._related = list(dict.fromkeys(main._related))
			if main._uuid != item._uuid:
				self.debug(f'found {len(others)} duplicates for', obj=item.toDict(), obj_breaklines=True, sub='duplicates', verbose=True)  # noqa: E501
				item._duplicate = True
				item = self.run_hooks('on_item', item)
				if item._uuid not in main._related:
					main._related.append(item._uuid)
				main = self.run_hooks('on_duplicate', main)
				item = self.run_hooks('on_duplicate', item)

			for dupe in dupes:
				if not dupe._duplicate:
					self.debug(
						'found new duplicate', obj=dupe.toDict(), obj_breaklines=True,
						sub='duplicates', verbose=True)
					# dupe_count += 1
					dupe._duplicate = True
					dupe = self.run_hooks('on_duplicate', dupe)

	def yielder(self):
		"""Yield results. Should be implemented by derived classes."""
		raise NotImplementedError()

	def yielder_celery(self):
		"""Yield results from Celery result."""
		yield from CeleryData.iter_results(
			self.celery_result,
			ids_map=self.celery_ids_map,
			print_remote_info=False
		)

	def toDict(self):
		"""Dict representation of the runner."""
		data = {
			'name': self.name,
			'status': self.status,
			'targets': self.inputs,
			'start_time': self.start_time,
			'end_time': self.end_time,
			'elapsed': self.elapsed.total_seconds(),
			'elapsed_human': self.elapsed_human,
			'run_opts': {k: v for k, v in self.run_opts.items() if k not in self.print_opts},
		}
		data.update({
			'config': self.config.toDict(),
			'opts': self.config.supported_opts,
			'has_parent': self.has_parent,
			'has_children': self.has_children,
			'chunk': self.chunk,
			'chunk_count': self.chunk_count,
			'sync': self.sync,
			'done': self.done,
			'output': self.output,
			'progress': self.progress,
			'last_updated_db': self.last_updated_db,
			'context': self.context,
			'errors': [e.toDict() for e in self.errors],
		})
		return data

	def run_hooks(self, hook_type, *args):
		""""Run hooks of a certain type.

		Args:
			hook_type (str): Hook type.
			args (list): List of arguments to pass to the hook.

		Returns:
			any: Hook return value.
		"""
		result = args[0] if len(args) > 0 else None
		_id = self.context.get('task_id', '') or self.context.get('workflow_id', '') or self.context.get('scan_id', '')
		for hook in self.hooks[hook_type]:
			name = f'{self.__class__.__name__}.{hook_type}'
			fun = self.get_func_path(hook)
			try:
				if hook_type == 'on_interval' and not should_update(CONFIG.runners.backend_update_frequency, self.last_updated_db):
					self.debug('', obj={f'{name} [dim yellow]->[/] {fun}': '[dim gray11]skipped[/]'}, id=_id, sub='hooks', verbose=True)  # noqa: E501
					return
				if not self.enable_hooks or self.no_process:
					self.debug('', obj={f'{name} [dim yellow]->[/] {fun}': '[dim gray11]skipped[/]'}, id=_id, sub='hooks', verbose=True)  # noqa: E501
					continue
				# self.debug('', obj={f'{name} [dim yellow]->[/] {fun}': '[dim yellow]started[/]'}, id=_id, sub='hooks', verbose=True)  # noqa: E501
				result = hook(self, *args)
				self.debug('', obj={f'{name} [dim yellow]->[/] {fun}': '[dim green]success[/]'}, id=_id, sub='hooks', verbose=True)  # noqa: E501
			except Exception as e:
				self.debug('', obj={f'{name} [dim yellow]->[/] {fun}': '[dim red]failed[/]'}, id=_id, sub='hooks', verbose=True)  # noqa: E501
				error = Error.from_exception(e)
				error.message = f'Hook "{fun}" execution failed.'
				error._source = self.unique_name
				error._uuid = str(uuid.uuid4())
				self.add_result(error, print=True)
				if self.raise_on_error:
					raise e
		return result

	def run_validators(self, validator_type, *args, error=True):
		"""Run validators of a certain type.

		Args:
			validator_type (str): Validator type. E.g: on_start.
			args (list): List of arguments to pass to the validator.
			error (bool): Whether to add an error to runner results if the validator failed.

		Returns:
			bool: Validator return value.
		"""
		if self.no_process:
			return True
		_id = self.context.get('task_id', '') or self.context.get('workflow_id', '') or self.context.get('scan_id', '')
		for validator in self.validators[validator_type]:
			name = f'{self.__class__.__name__}.{validator_type}'
			fun = self.get_func_path(validator)
			if not validator(self, *args):
				self.debug('', obj={name + ' [dim yellow]->[/] ' + fun: '[dim red]failed[/]'}, id=_id, verbose=True, sub='validators')  # noqa: E501
				doc = validator.__doc__
				if error:
					message = 'Validator failed'
					if doc:
						message += f': {doc}'
					error = Error(
						message=message,
						_source=self.unique_name,
						_uuid=str(uuid.uuid4())
					)
					self.add_result(error, print=True)
				return False
			self.debug('', obj={name + ' [dim yellow]->[/] ' + fun: '[dim green]success[/]'}, id=_id, verbose=True, sub='validators')  # noqa: E501
		return True

	def register_hooks(self, hooks):
		"""Register hooks.

		Args:
			hooks (dict[str, List[Callable]]): List of hooks to register.
		"""
		for key in self.hooks:
			# Register class + derived class hooks
			class_hook = getattr(self, key, None)
			if class_hook:
				name = f'{self.__class__.__name__}.{key}'
				fun = self.get_func_path(class_hook)
				self.debug('', obj={name + ' [dim yellow]->[/] ' + fun: 'registered'}, sub='hooks')
				self.hooks[key].append(class_hook)

			# Register user hooks
			user_hooks = hooks.get(self.__class__, {}).get(key, [])
			user_hooks.extend(hooks.get(key, []))
			for hook in user_hooks:
				name = f'{self.__class__.__name__}.{key}'
				fun = self.get_func_path(hook)
				self.debug('', obj={name + ' [dim yellow]->[/] ' + fun: 'registered (user)'}, sub='hooks')
			self.hooks[key].extend(user_hooks)

	def register_validators(self, validators):
		"""Register validators.

		Args:
			validators (dict[str, List[Callable]]): Validators to register.
		"""
		# Register class + derived class hooks
		for key in self.validators:
			class_validator = getattr(self, key, None)
			if class_validator:
				name = f'{self.__class__.__name__}.{key}'
				fun = self.get_func_path(class_validator)
				self.validators[key].append(class_validator)
				self.debug('', obj={name + ' [dim yellow]->[/] ' + fun: 'registered'}, sub='validators')

			# Register user hooks
			user_validators = validators.get(key, [])
			for validator in user_validators:
				name = f'{self.__class__.__name__}.{key}'
				fun = self.get_func_path(validator)
				self.debug('', obj={name + ' [dim yellow]->[/] ' + fun: 'registered (user)'}, sub='validators')
			self.validators[key].extend(user_validators)

	def log_start(self):
		"""Log runner start."""
		if not self.print_remote_info:
			return
		remote_str = 'starting' if self.sync else 'sent to Celery worker'
		runner_name = self.__class__.__name__
		info = Info(message=f'{runner_name} {self.config.name} {remote_str}...', _source=self.unique_name)
		self._print_item(info)

	def log_results(self):
		"""Log runner results."""
		if self.no_poll:
			return
		self.done = True
		self.progress = 100
		self.end_time = datetime.fromtimestamp(time())
		if self.status == 'FAILURE':
			self.debug('', obj={self.__class__.__name__: self.status, 'errors': [str(_.message) for _ in self.errors]}, sub='status')  # noqa: E501
		else:
			self.debug('', obj={self.__class__.__name__: self.status}, sub='status')
		if self.exporters and not self.no_process:
			report = Report(self, exporters=self.exporters)
			report.build()
			report.send()
			self.report = report
		if self.enable_profiler:
			self.profiler.stop()
			profile_path = Path(self.reports_folder) / f'{self.unique_name}_profile.html'
			with profile_path.open('w', encoding='utf-8') as f_html:
				f_html.write(self.profiler.output_html())
			self._print_item(Info(message=f'Wrote profile to {str(profile_path)}', _source=self.unique_name), force=True)

	def stop_celery_tasks(self):
		"""Stop all tasks running in Celery worker."""
		from secator.celery import revoke_task
		for task_id in self.celery_ids:
			name = self.celery_ids_map.get(task_id, {}).get('full_name')
			revoke_task(task_id, name)

	def filter_results(self):
		"""Filter runner results using extractors defined in config."""
		extractors = self.config.results
		results = []
		if extractors:
			# Keep results based on extractors
			opts = merge_opts(self.config.options, self.run_opts)
			for extractor in extractors:
				tmp = process_extractor(self.results, extractor, ctx=opts)
				results.extend(tmp)

			# Keep the field types in results not specified in the extractors.
			extract_fields = [e['type'] for e in extractors]
			keep_fields = [
				_type for _type in FINDING_TYPES
				if _type not in extract_fields
			]
			results.extend([
				item for item in self.results
				if item._type in keep_fields
			])
		else:
			results = self.results
		return results

	def _convert_item_schema(self, item):
		"""Convert dict item to a secator output type.

		Args:
			item (dict): Dict item.

		Returns:
			OutputType: Loaded item.
		"""
		# Skip if already converted
		if isinstance(item, OutputType):
			return item

		# Init the new item and the list of output types to load from
		new_item = None
		output_types = getattr(self, 'output_types', [])
		self.debug(f'Input item: {item}', sub='klass.load', verbose=True)

		# Use a function to pick proper output types
		output_discriminator = getattr(self, 'output_discriminator', None)
		if output_discriminator:
			result = output_discriminator(item)
			if result:
				self.debug(f'Discriminated output type: {result.__name__}', sub='klass.load', verbose=True)
				output_types = [result]
			else:
				output_types = []

		# Use the _type key to pick proper output type
		elif '_type' in item:
			otypes = [o for o in output_types if o.get_name() == item['_type']]
			if otypes:
				output_types = [otypes[0]]
				self.debug(f'_type key is present in item and matches {otypes[0]}', sub='klass.load', verbose=True)

		# Load item using picked output types
		self.debug(f'Output types to try: {[o.__name__ for o in output_types]}', sub='klass.load', verbose=True)
		for klass in output_types:
			self.debug(f'Loading item as {klass.__name__}', sub='klass.load', verbose=True)
			output_map = getattr(self, 'output_map', {}).get(klass, {})
			try:
				new_item = klass.load(item, output_map)
				self.debug(f'[dim green]Successfully loaded item as {klass.__name__}[/]', sub='klass.load', verbose=True)
				break
			except (TypeError, KeyError) as e:
				self.debug(
					f'[dim red]Failed loading item as {klass.__name__}: {type(e).__name__}: {str(e)}.[/] [dim green]Continuing.[/]',
					sub='klass.load', verbose=True)
				# error = Error.from_exception(e)
				# self.debug(repr(error), sub='debug.klass.load')
				continue

		if not new_item:
			new_item = Warning(message=f'Failed to load item as output type:\n  {item}')

		self.debug(f'Output item: {new_item.toDict()}', sub='klass.load', verbose=True)

		return new_item

	def _print(self, data, color=None, out=sys.stderr, rich=False, end='\n'):
		"""Print function.

		Args:
			data (str or dict): Input data.
			color (str, Optional): Rich color.
			out (str, Optional): Output pipe (sys.stderr, sys.stdout, ...)
			rich (bool, Optional): Force rich output.
			end (str, Optional): End of line.
			add_to_output (bool, Optional): Whether to add the item to runner output.
		"""
		if rich:
			_console = console_stdout if out == sys.stdout else console
			_console.print(data, highlight=False, style=color, soft_wrap=True, end=end)
		else:
			if isinstance(data, (OutputType, dict)):
				if getattr(data, 'toDict', None):
					data = data.toDict()
				data = json.dumps(data)
			print(data, file=out)

	def _get_findings_count(self):
		"""Get finding count.

		Returns:
			dict[str,int]: Dict of finding type to count.
		"""
		count_map = {}
		for output_type in FINDING_TYPES:
			name = output_type.get_name()
			count = len([r for r in self.results if isinstance(r, output_type)])
			if count > 0:
				count_map[name] = count
		return count_map

	def _process_item(self, item, print=True, output=True):
		"""Process an item yielded by the derived runner.

		Args:
			item (dict | str): Input item.
			print (bool): Print item in console.
			output (bool): Add to runner output.

		Yields:
			OutputType: Output type.
		"""
		# Item is a string, just print it
		if isinstance(item, str):
			self.output += item + '\n' if output else ''
			self._print_item(item) if item and print else ''
			return

		# Abort further processing if no_process is set
		if self.no_process:
			return

		# Run item validators
		if not self.run_validators('validate_item', item, error=False):
			return

		# Convert output dict to another schema
		if isinstance(item, dict):
			item = self.run_hooks('on_item_pre_convert', item)
			if not item:
				return
			item = self._convert_item_schema(item)

		# Update item context
		item._context.update(self.context)

		# Return if already seen
		if item._uuid in self.uuids:
			return

		# Add uuid to item
		if not item._uuid:
			item._uuid = str(uuid.uuid4())

		# Add source to item
		if not item._source:
			item._source = self.unique_name

		# If progress item, update runner progress
		if isinstance(item, Progress) and item._source == self.unique_name:
			self.progress = item.percent
			if not should_update(CONFIG.runners.progress_update_frequency, self.last_updated_progress, item._timestamp):
				return
			elif int(item.percent) in [0, 100]:
				return
			else:
				self.last_updated_progress = item._timestamp

		# If info item and task_id is defined, update runner celery_ids
		elif isinstance(item, Info) and item.task_id and item.task_id not in self.celery_ids:
			self.celery_ids.append(item.task_id)

		# If finding, run on_item hooks
		elif isinstance(item, tuple(FINDING_TYPES)):
			item = self.run_hooks('on_item', item)
			if not item:
				return

		# Add item to results
		self.add_result(item, print=print)

		# Yield item
		yield item

	@staticmethod
	def resolve_exporters(exporters):
		"""Resolve exporters from output options.

		Args:
			exporters (list[str]): List of exporters to resolve.

		Returns:
			list: List of exporter classes.
		"""
		if not exporters or exporters in ['false', 'False']:
			return []
		if isinstance(exporters, str):
			exporters = exporters.split(',')
		classes = [
			import_dynamic('secator.exporters', f'{o.capitalize()}Exporter')
			for o in exporters
			if o
		]
		return [cls for cls in classes if cls]

	@classmethod
	def get_func_path(cls, func):
		"""Get the full symbolic path of a function or method, including staticmethods, using function and method
		attributes.

		Args:
			func (function, method, or staticmethod): A function or method object.
		"""
		if hasattr(func, '__self__'):
			if func.__self__ is not None:
				# It's a method bound to an instance
				class_name = func.__self__.__class__.__name__
				return f"{func.__module__}.{class_name}.{func.__name__}"
			else:
				# It's a method bound to a class (class method)
				class_name = func.__qualname__.rsplit('.', 1)[0]
				return f"{func.__module__}.{class_name}.{func.__name__}"
		else:
			# Handle static and regular functions
			if '.' in func.__qualname__:
				# Static method or a function defined inside a class
				class_name, func_name = func.__qualname__.rsplit('.', 1)
				return f"{func.__module__}.{class_name}.{func_name}"
			else:
				# Regular function not attached to a class
				return f"{func.__module__}.{func.__name__}"



from secator.decorators import task
from secator.definitions import (DOMAIN, HOST, RATE_LIMIT, RETRIES, THREADS, WORDLIST, EXTRA_DATA)
from secator.config import CONFIG
from secator.output_types import Subdomain
from secator.serializers import JSONSerializer
from secator.tasks._categories import ReconDns
from secator.utils import process_wordlist


@task()
class dnsxbrute(ReconDns):
    """dnsx is a fast and multi-purpose DNS toolkit designed for running various library."""
    cmd = 'dnsx'
    json_flag = '-json'
    input_flag = '-domain'
    file_flag = '-domain'
    opt_key_map = {
        RATE_LIMIT: 'rate-limit',
        RETRIES: 'retry',
        THREADS: 'threads',
    }
    opts = {
        WORDLIST: {'type': str, 'short': 'w', 'default': CONFIG.wordlists.defaults.dns, 'process': process_wordlist, 'help': 'Wordlist to use'},  # noqa: E501
        'trace': {'is_flag': True, 'default': False, 'help': 'Perform dns tracing'},
    }
    item_loaders = [JSONSerializer()]
    output_map = {
        Subdomain: {
            HOST: 'host',
            DOMAIN: lambda x: ".".join(x['host'].split('.')[1:]),
            EXTRA_DATA: lambda x: {
                'resolver': x['resolver'],
                'status_code': x['status_code']
			}
        }
    }
    install_cmd = 'go install -v github.com/projectdiscovery/dnsx/cmd/dnsx@latest'
    install_github_handle = 'projectdiscovery/dnsx'
    profile = 'io'



import re


class RegexSerializer:

	def __init__(self, regex, fields=[], findall=False):
		self.regex = re.compile(regex)
		self.fields = fields
		self.findall = findall

	def run(self, line):
		if self.findall:
			match = self.regex.findall(line)
			yield from match
			return
		output = {}
		match = self.regex.match(line)
		if not match:
			return
		if not self.fields:
			yield match.group(0)
			return
		for field in self.fields:
			output[field] = match.group(field)
		yield output



import eventlet
eventlet.monkey_patch()
from secator.runners import Workflow, Task, Scan
from secator.template import TemplateLoader
from secator.rich import console
from secator.celery import *
import click
import sys
import json
from time import sleep, time

pool = eventlet.GreenPool(100)

from kombu.serialization import register

def create_runner(runner_type, targets, index, total):
	register('json', json.dumps, json.loads, content_type='application/json', content_encoding='utf-8')
	run_opts = {
		'print_item': True,
		'print_line': True,
		'print_cmd': True
	}
	runner = None
	if runner_type == 'workflow':
		runner = Workflow
		config = TemplateLoader(name='workflows/subdomain_recon')
	elif runner_type == 'task':
		runner = Task
		config = TemplateLoader(input={'name': 'httpx', 'type': 'task'})
	elif runner_type == 'scan':
		runner = Scan
	result = runner.delay(config, targets, run_opts=run_opts)
	while not result.ready():
		# print(f'Running {runner_type} {index}/{total} ..')
		sleep(1)
	print(f'Task {index} / {total} finished running.')
	result.get()

def start_worker():
	from secator.runners import Command
	from threading import Thread
	cmd = f'pyinstrument -r html -o /tmp/test.html --from-path secator worker'
	process = Command.execute(cmd, run=False)
	thread = Thread(target=process.run)
	return thread, process

@click.command()
@click.option('--count', type=int, default=1)
@click.option('--runner', type=str, default='task')
@click.option('--targets', type=str, default='http://testphp.vulnweb.com')
def cli(count, runner, targets):
	"""Secator CLI."""
	try:
		targets = [c.strip() for c in targets.split(',')]
		print(f'Load tester initialized with {len(targets)} targets.')
		print(f'Targets: {targets}')
		thread, process = start_worker()
		print('Starting worker ...')
		thread.start()
		sleep(3)
		print(f'Starting {count} {runner}s ...')
		start_time = time()
		for i in range(count):
			print(f'Starting {runner} {i + 1}/{count} ..')
			pool.spawn(create_runner, runner, targets, i + 1, count)
		pool.waitall()
		elapsed_time = time() - start_time
		print(f'All {runner}s completed in {elapsed_time:.2f}s.')
	except BaseException:
		pass
	finally:
		import signal
		process.send_signal(signal.SIGINT)
		thread.join()

if __name__ == '__main__':
    cli()



from secator.decorators import task
from secator.definitions import (DELAY, DEPTH, FILTER_CODES, FILTER_REGEX,
							   FILTER_SIZE, FILTER_WORDS, FOLLOW_REDIRECT,
							   HEADER, MATCH_CODES, MATCH_REGEX, MATCH_SIZE,
							   MATCH_WORDS, METHOD, OPT_NOT_SUPPORTED,
							   OPT_PIPE_INPUT, PROXY, RATE_LIMIT, RETRIES,
							   THREADS, TIMEOUT, URL, USER_AGENT)
from secator.output_types import Tag, Url
from secator.serializers import JSONSerializer
from secator.tasks._categories import HttpCrawler


@task()
class cariddi(HttpCrawler):
	"""Crawl endpoints, secrets, api keys, extensions, tokens..."""
	cmd = 'cariddi -info -s -err -e -ext 1'
	input_type = URL
	input_flag = OPT_PIPE_INPUT
	output_types = [Url, Tag]
	file_flag = OPT_PIPE_INPUT
	json_flag = '-json'
	opt_key_map = {
		HEADER: 'headers',
		DELAY: 'd',
		DEPTH: OPT_NOT_SUPPORTED,
		FILTER_CODES: OPT_NOT_SUPPORTED,
		FILTER_REGEX: OPT_NOT_SUPPORTED,
		FILTER_SIZE: OPT_NOT_SUPPORTED,
		FILTER_WORDS: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		MATCH_CODES: OPT_NOT_SUPPORTED,
		MATCH_REGEX: OPT_NOT_SUPPORTED,
		MATCH_SIZE: OPT_NOT_SUPPORTED,
		MATCH_WORDS: OPT_NOT_SUPPORTED,
		METHOD: OPT_NOT_SUPPORTED,
		PROXY: 'proxy',
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: OPT_NOT_SUPPORTED,
		THREADS: 'c',
		TIMEOUT: 't',
		USER_AGENT: 'ua'
	}
	item_loaders = [JSONSerializer()]
	install_cmd = 'go install -v github.com/edoardottt/cariddi/cmd/cariddi@latest'
	install_github_handle = 'edoardottt/cariddi'
	encoding = 'ansi'
	proxychains = False
	proxy_socks5 = True  # with leaks... https://github.com/edoardottt/cariddi/issues/122
	proxy_http = True  # with leaks... https://github.com/edoardottt/cariddi/issues/122
	profile = 'cpu'

	@staticmethod
	def on_json_loaded(self, item):
		url_item = {k: v for k, v in item.items() if k != 'matches'}
		yield Url(**url_item)
		url = url_item[URL]
		matches = item.get('matches', {})
		params = matches.get('parameters', [])
		errors = matches.get('errors', [])
		secrets = matches.get('secrets', [])
		infos = matches.get('infos', [])

		for param in params:
			param_name = param['name']
			for attack in param['attacks']:
				extra_data = {'param': param_name, 'source': 'url'}
				yield Tag(
					name=f'{attack} param',
					match=url,
					extra_data=extra_data
				)

		for error in errors:
			match = error['match']
			error['extra_data'] = {'error': match, 'source': 'body'}
			error['match'] = url
			yield Tag(**error)

		for secret in secrets:
			match = secret['match']
			secret['extra_data'] = {'secret': match, 'source': 'body'}
			secret['match'] = url
			yield Tag(**secret)

		for info in infos:
			CARIDDI_IGNORE_LIST = ['BTC address']  # TODO: make this a config option
			if info['name'] in CARIDDI_IGNORE_LIST:
				continue
			match = info['match']
			info['extra_data'] = {'info': match, 'source': 'body'}
			info['match'] = url
			yield Tag(**info)



import json
import os

from secator.config import CONFIG
from secator.decorators import task
from secator.definitions import (CONFIDENCE, CVSS_SCORE, DELAY, DESCRIPTION,
							   EXTRA_DATA, FOLLOW_REDIRECT, HEADER, ID,
							   MATCHED_AT, NAME, OPT_NOT_SUPPORTED, OUTPUT_PATH, PROVIDER,
							   PROXY, RATE_LIMIT, REFERENCES, RETRIES,
							   SEVERITY, TAGS, THREADS, TIMEOUT,
							   URL, USER_AGENT)
from secator.output_types import Tag, Vulnerability, Info, Error
from secator.tasks._categories import VulnHttp


@task()
class wpscan(VulnHttp):
	"""Wordpress security scanner."""
	cmd = 'wpscan --random-user-agent --force --verbose'
	file_flag = None
	input_flag = '--url'
	input_type = URL
	json_flag = '-f json'
	opt_prefix = '--'
	opts = {
		'cookie_string': {'type': str, 'short': 'cookie', 'help': 'Cookie string, format: cookie1=value1;...'},
		'api_token': {'type': str, 'short': 'token', 'help': 'WPScan API Token to display vulnerability data'},
		'wp_content_dir': {'type': str, 'short': 'wcd', 'help': 'wp-content directory if custom or not detected'},
		'wp_plugins_dir': {'type': str, 'short': 'wpd', 'help': 'wp-plugins directory if custom or not detected'},
		'passwords': {'type': str, 'help': 'List of passwords to use during the password attack.'},
		'usernames': {'type': str, 'help': 'List of usernames to use during the password attack.'},
		'login_uri': {'type': str, 'short': 'lu', 'help': 'URI of the login page if different from /wp-login.php'},
		'detection_mode': {'type': str, 'short': 'dm', 'help': 'Detection mode between mixed, passive, and aggressive'}
	}
	opt_key_map = {
		HEADER: OPT_NOT_SUPPORTED,
		DELAY: 'throttle',
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		PROXY: 'proxy',
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: OPT_NOT_SUPPORTED,
		TIMEOUT: 'request-timeout',
		THREADS: 'max-threads',
		USER_AGENT: 'user-agent',
	}
	opt_value_map = {
		DELAY: lambda x: x * 1000
	}
	output_map = {
		Vulnerability: {
			ID: lambda x: '',
			NAME: lambda x: x['to_s'].split(':')[0],
			DESCRIPTION: lambda x: '',
			SEVERITY: lambda x: 'info',
			CONFIDENCE: lambda x: 'high' if x.get('confidence', 0) == 100 else 'low',
			CVSS_SCORE: lambda x: 0,
			MATCHED_AT: lambda x: x['url'],
			TAGS: lambda x: [x['type']],
			REFERENCES: lambda x: x.get('references', {}).get('url', []),
			EXTRA_DATA: lambda x: {
				'data': x.get('interesting_entries', []),
				'found_by': x.get('found_by', ''),
				'confirmed_by': x.get('confirmed_by', {}),
				'metasploit': x.get('references', {}).get('metasploit', [])
			},
			PROVIDER: 'wpscan',
		},
	}
	output_types = [Vulnerability, Tag]
	install_pre = {
		'apt': ['make', 'kali:libcurl4t64', 'libffi-dev'],
		'pacman': ['make', 'ruby-erb'],
		'*': ['make']
	}
	install_cmd = f'gem install wpscan --user-install -n {CONFIG.dirs.bin}'
	install_post = {
		'kali': (
			f'gem uninstall nokogiri --user-install -n {CONFIG.dirs.bin} --force --executables && '
			f'gem install nokogiri --user-install -n {CONFIG.dirs.bin} --platform=ruby'
		)
	}
	proxychains = False
	proxy_http = True
	proxy_socks5 = False
	profile = 'io'

	@staticmethod
	def on_init(self):
		output_path = self.get_opt_value(OUTPUT_PATH)
		if not output_path:
			output_path = f'{self.reports_folder}/.outputs/{self.unique_name}.json'
		self.output_path = output_path
		self.cmd += f' -o {self.output_path}'

	@staticmethod
	def on_cmd_done(self):
		if not os.path.exists(self.output_path):
			yield Error(message=f'Could not find JSON results in {self.output_path}')
			return

		yield Info(message=f'JSON results saved to {self.output_path}')
		with open(self.output_path, 'r') as f:
			data = json.load(f)

		# Get URL
		target = data.get('target_url', self.inputs[0])

		# Wordpress version
		version = data.get('version', {})
		if version:
			wp_version = version['number']
			wp_version_status = version['status']
			if wp_version_status == 'outdated':
				vuln = version
				vuln.update({
					'url': target,
					'to_s': 'Wordpress outdated version',
					'type': wp_version,
					'references': {},
				})
				yield vuln

		# Main theme
		main_theme = data.get('main_theme', {})
		if main_theme:
			version = main_theme.get('version', {})
			slug = main_theme['slug']
			location = main_theme['location']
			if version:
				number = version['number']
				latest_version = main_theme.get('latest_version')
				yield Tag(
					name=f'Wordpress theme - {slug} {number}',
					match=target,
					extra_data={
						'url': location,
						'latest_version': latest_version
					}
				)
				if (latest_version and number < latest_version):
					yield Vulnerability(
						matched_at=target,
						name=f'Wordpress theme - {slug} {number} outdated',
						confidence='high',
						severity='info'
					)

		# Interesting findings
		interesting_findings = data.get('interesting_findings', [])
		for item in interesting_findings:
			yield item

		# Plugins
		plugins = data.get('plugins', {})
		for _, data in plugins.items():
			version = data.get('version', {})
			slug = data['slug']
			location = data['location']
			if version:
				number = version['number']
				latest_version = data.get('latest_version')
				yield Tag(
					name=f'Wordpress plugin - {slug} {number}',
					match=target,
					extra_data={
						'url': location,
						'latest_version': latest_version
					}
				)
				if (latest_version and number < latest_version):
					yield Vulnerability(
						matched_at=target,
						name=f'Wordpress plugin - {slug} {number} outdated',
						confidence='high',
						severity='info'
					)
